# # Comparative Model Validation

# # Comparative Model Validation

# imports
import os
import feather
import pandas
from IPython.display import display
import pickle
import sklearn.metrics as error_metrics

# + tags=["parameters"]
upstream = None
product = None
models = []
split_type = ''
n_folds = 10
quantile_intervals = []
# -

# format variable
upstream = dict(upstream)

# ## Load predictions and ground truth

# load ground truth data from upstream
if split_type == 'train_test':
    ground_truth = feather.read_dataframe(upstream['split_data']['test_y_data']).reset_index(drop=True)
elif split_type == 'k_fold':
    ground_truth = []
    for i in range(1, n_folds+1):   
        ground_truth.append(
            feather.read_dataframe(upstream['split_data'][f'fold_{i}_y_data'])
        )
    ground_truth = pandas.concat(ground_truth).reset_index(drop=True)
del upstream['split_data']

# load predictions from upstream
predictions = {}
for model_name in models:
    predictions[model_name] = feather.read_dataframe(
        upstream[f'{model_name}_predict']['predictions']
    ).reset_index(drop=True)

# ## Error Metrics
# The error metrics reported are
# * MSE - Mean squared error
# * RMSE - Root mean squared error
# * MAE - Mean absolute error
# * MAPE - Mean absolute percentage error
# * R-squared - Coefficient of determination

# error
error_df = pandas.DataFrame()
# mse
for model, preds in predictions.items():
    mean_preds = preds['predictions']
    error_df.loc[model, 'MSE'] = error_metrics.mean_squared_error(ground_truth, mean_preds)
    error_df.loc[model, 'RMSE'] = error_metrics.mean_squared_error(ground_truth, mean_preds, squared=False)
    error_df.loc[model, 'MAE'] = error_metrics.mean_absolute_error(ground_truth, mean_preds)
    error_df.loc[model, 'MAPE'] = error_metrics.mean_absolute_percentage_error(ground_truth, mean_preds)
    error_df.loc[model, 'R-squared'] = error_metrics.r2_score(ground_truth, mean_preds)
error_df

# ## Quantile interval coverage
# For each of the intervals defined in `quantile_intervals`, what proportion of predicted values fall within the interval?

# quantile interval coverage
qi_coverage = pandas.DataFrame()
idx = 0
for model, preds in predictions.items():
    for qi in quantile_intervals:
        lb = 'Q' + f'{min(qi) * 100:g}'
        ub = 'Q' + f'{max(qi) * 100:g}'
        if lb in preds.columns and ub in preds.columns:
            qi_coverage.loc[idx, 'model'] = model
            qi_coverage.loc[idx, 'interval'] = str(sorted(qi))
            qi_coverage.loc[idx, 'level'] = max(qi) - min(qi)
            # boolean array for qi covering ground truth
            in_interval = (preds[ub] >= ground_truth.iloc[:, 0]) & (preds[lb] <= ground_truth.iloc[:, 0])
            qi_coverage.loc[idx, 'coverage'] = in_interval.sum() / in_interval.shape[0]
            idx += 1
try:
    display(qi_coverage.set_index(['model', 'interval']))
except KeyError:
    print('No quantile intervals specified. Nothing to calculate.')

# ## Model Sharpness
# How uncertain is the model? `sharpness` of a quantile interval is defined as the average width of the interval across all predictions.

# quantile interval sharpness
qi_sharpness = pandas.DataFrame()
for model, preds in predictions.items():
    for qi in quantile_intervals:
        lb = 'Q' + f'{min(qi) * 100:g}'
        ub = 'Q' + f'{max(qi) * 100:g}'
        if lb in preds.columns and ub in preds.columns:
            qi_sharpness.loc[model, str(sorted(qi))] = (preds[ub] - preds[lb]).mean()
qi_sharpness

# ## Computational performance
# Record
# * the total size in bytes of the fit model(s).
# * the total time taken to fit the model(s).
# * the total time taken for the model(s) to predict.

# computational performance
comp_perf = pandas.DataFrame()
for model in models:
    # fit time
    with open(upstream[f'{model}_fit']['process_metadata'], 'rb') as f:
        fit_metadata = pickle.load(f)
    comp_perf.loc[model, 'fit time'] = fit_metadata[f'{model}_fit']
    # predict time
    with open(upstream[f'{model}_predict']['process_metadata'], 'rb') as f:
        predict_metadata = pickle.load(f)
    comp_perf.loc[model, 'predict time'] = predict_metadata[f'{model}_predict']
    # model size
    if split_type == 'train_test':
        # only one model
        comp_perf.loc[model, 'size'] = os.path.getsize(
            os.path.join(
                'output',
                'models',
                f'{model}_fit'
            )
        )
    elif split_type == 'k_fold':
        # sum the model sizes
        total_size = 0
        for i in range(1, n_folds+1):
            total_size += os.path.getsize(
                os.path.join(
                    'output',
                    'models',
                    f'{model}_fit_{i}' 
                )
            )
        comp_perf.loc[model, 'size'] = total_size
comp_perf


# R models file

# MVF imports
library(R6)
# User imports
{% for model in models %}

{{model}} <- R6Class(
    # Class implementing {{model}}.
    '{{model}}',
    public = list(
        model = NULL,
        fit = function(X, y){
            # Fit model to data.
            .NotYetImplemented()
        },
        predict = function(X, quantile_intervals){
            # Predict target values (y) from new data.
            .NotYetImplemented()
        },
        save = function(path){
            # Save model.
            .NotYetImplemented()
        },
        load = function(path){
            # Load a model.
            .NotYetImplemented()
        }
    )
)
{% endfor %}


# Python models file

# MVF imports
import pandas
# User imports
{% for model in models %}

class {{model}}:
    '''
    Class implementing {{model}}.
    '''

    def fit(self, X, y):
        '''
        Fit model to data.
        '''
        raise NotImplementedError

    def predict(self, X, quantiles=None):
        '''
        Predict target values (y) from new data.
        '''
        raise NotImplementedError

    def save(self, path):
        '''
        Save model.
        '''
        raise NotImplementedError

    def load(self, path):
        '''
        Load a model.
        '''
        raise NotImplementedError
{% endfor %}



# imports
import feather
import pandas
from mvf.process.utils import ProcessTimer, convert_to_pandas_data, import_model_class, init_model, convert_to_r_data


def predict_model(upstream, product, lang, model_name, split_type, n_folds=10, quantile_intervals=None):
    # start process timer
    timer = ProcessTimer(f'{model_name}_predict')
    # format variable
    upstream = dict(upstream)

    # import model class
    model_module = import_model_class(lang)

    if split_type == 'train_test':
        preds = make_prediction(
            upstream['split_data']['test_X_data'],
            str(upstream[f'{model_name}_fit']['model']),
            lang,
            model_module,
            model_name,
            quantile_intervals
        )
        # save data for next process
        feather.write_dataframe(preds, product['predictions'])
    elif split_type == 'k_fold':
        # allocate memory for predictions
        predictions = []
        # for each fold
        for i in range(1, n_folds+1):
            preds = make_prediction(
                upstream['split_data'][f'fold_{i}_X_data'],
                str(upstream[f'{model_name}_fit'][f'model_{i}']),
                lang,
                model_module,
                model_name,
                quantile_intervals
            )
            # append fold predictions to predictions set
            predictions.append(preds)
        # save data for next process
        feather.write_dataframe(
            pandas.concat(predictions),
            product['predictions']
        )
    # end process timer
    timer.end()
    # save process metadata
    timer.save(product['process_metadata'])
    

def make_prediction(data_path, model_path, lang, model_module, model_name, quantile_intervals):
    '''
    Makes a prediction based on a set of test data and a path to a fit model.
    '''
    # load test data
    X_test = feather.read_dataframe(data_path)

    # initialise and load model
    model = init_model(
        model_module,
        lang,
        model_name
    )
    model.load(model_path)

    if lang == 'R':
        # convert pandas dataframe to R dataframe
        X_test = convert_to_r_data(X_test)
        if quantile_intervals is None:
            # convert quantiles to R data
            r = model_module['r']
            quantile_intervals = r('NULL')

    # predict
    preds = model.predict(X_test, quantile_intervals)

    if lang == 'R':
        # convert to pandas dataframe
        preds = convert_to_pandas_data(preds)
    return preds




import click
import os
import distutils.dir_util
import subprocess
import yaml
from mvf.integration import mvf_config


def load_config(path):
    '''
    Loads the mvf_conf.yaml from the working directory 
    and validates the config against the schema.
    '''
    # open the mvf config file
    try:
        with open(path, 'r') as f:
            config = yaml.safe_load(f)
    except FileNotFoundError:
        raise Exception('No `mvf_conf.yaml` found in the working directory.')
    else:
        # validate the config against the schema
        click.echo('Validating config...')
        mvf_config.check_config(config)
        return config


def download_example(example_name, output_dir):
    '''
    Downloads an example MVF project.
    '''
    branch = 'main'
    url = 'https://gitlab.com/TomKimCTA/model-validation-framework'
    example_path = f'test/test_resources/test_{example_name}'
    # Check target directory does not already exist
    if os.path.isdir(output_dir):
        raise Exception(f'{output_dir} is already a directory.')
    # Create sparse repository
    subprocess.run(
        [
            'git',
            'clone',
            '-n',
            '--depth',
            '1',
            '-b',
            branch,
            '--filter=blob:none',
            url,
            output_dir
        ]
    )
    os.chdir(output_dir)
    # Set which subfolder to checkout
    subprocess.run(
        [
            'git',
            'sparse-checkout',
            'set',
            example_path
        ]
    )
    # Checkout example
    subprocess.run(
        [
            'git',
            'checkout'
        ]
    )
    # Reorganise directories
    distutils.dir_util.copy_tree(
        os.path.join(
            'test',
            'test_resources',
            f'test_{example_name}'
        ),
        os.getcwd()
    )
    # Clean up directory
    distutils.dir_util.remove_tree('test')



import timeit
import pickle


class ProcessTimer:
    def __init__(self, process_name):
        self.process_name = process_name
        self.start = timeit.default_timer()

    def end(self):
        self.end = timeit.default_timer()

    def save(self, path):
        self.process_time = {
            self.process_name: self.end - self.start
        }
        with open(path, 'wb') as f:
            pickle.dump(self.process_time, f, protocol=pickle.HIGHEST_PROTOCOL)


def import_model_class(lang):
    '''
    Imports model module depending on language parameter.
    '''
    if lang == 'Python':
        import models
        return models
    elif lang == 'R':
        import rpy2.robjects as robjects
        import rpy2_r6.r6b as r6b

        r = robjects.r
        r['source']('models.R')

        return {
            'r6b': r6b,
            'r': r
        }


def init_model(model_module, lang, model_name):
    '''
    Instantiates a model.
    '''
    if lang == 'Python':
        model_class = getattr(model_module, model_name)
        model = model_class()
    elif lang == 'R':
        r6b = model_module['r6b']
        r = model_module['r']
        model_class = r6b.R6DynamicClassGenerator(r[model_name])
        model = model_class.new()
    return model


def convert_to_r_data(data):
    '''
    Convert pandas.DataFrame to R data for use by rpy2.
    '''
    from rpy2.robjects import pandas2ri
    pandas2ri.activate()
    r_data = pandas2ri.py2rpy_pandasdataframe(
        data
    )
    pandas2ri.deactivate()
    return r_data


def convert_to_pandas_data(data):
    '''
    Convert R data from rpy2 to pandas.DataFrame.
    '''
    from rpy2.robjects import pandas2ri
    import rpy2.robjects as robjects
    pandas2ri.activate()
    pandas_data = robjects.conversion.rpy2py(data)
    pandas2ri.deactivate()
    return pandas_data